# Copyright (C) 2017 Roman Samoilenko <ttahabatt@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import asyncio
import unittest

from heralding.capabilities.vnc import Vnc, RFB_VERSION, VNC_AUTH
from heralding.reporting.reporting_relay import ReportingRelay


class VncTests(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

        self.reporting_relay = ReportingRelay()
        self.reporting_relay_task = self.loop.run_in_executor(None, self.reporting_relay.start)

    def tearDown(self):
        self.reporting_relay.stop()
        # We give reporting_relay a chance to be finished
        self.loop.run_until_complete(self.reporting_relay_task)

        self.server.close()
        self.loop.run_until_complete(self.server.wait_closed())

        self.loop.close()

    def test_vnc_authentication(self):
        async def vnc_auth():
            reader, writer = await asyncio.open_connection('127.0.0.1', 8888,
                                                           loop=self.loop)
            # server rfb version
            _ = await reader.readline()
            writer.write(RFB_VERSION)

            # available auth methods
            _ = await reader.read(1024)
            writer.write(VNC_AUTH)

            # challenge
            _ = await reader.read(1024)
            # Pretending, that we encrypt received challenge with DES and send back the result.
            client_response = os.urandom(16)
            writer.write(client_response)

            # security result
            _ = await reader.read(1024)

        options = {'enabled': 'True', 'port': 8888, 'timeout': 30}
        capability = Vnc(options, self.loop)

        server_coro = asyncio.start_server(capability.handle_session, '0.0.0.0', 8888, loop=self.loop)
        self.server = self.loop.run_until_complete(server_coro)

        self.loop.run_until_complete(vnc_auth())



# We need this in order to reduce the number of third-party modules.
# It is a part of code from http.client, that adjusted to work with
# asyncio in our specific case.

import email.parser
from http.client import HTTPMessage

async def parse_headers(fp, _class=HTTPMessage):
    headers = []
    while True:
        line = await fp.readline()
        headers.append(line)
        if line in (b'\r\n', b'\n', b''):
            break
    hstring = b''.join(headers).decode('iso-8859-1')
    return email.parser.Parser(_class=_class).parsestr(hstring)



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import logging

from heralding.misc.session import Session

logger = logging.getLogger(__name__)


class HandlerBase:
    MAX_GLOBAL_SESSIONS = 800
    global_sessions = 0

    def __init__(self, options, loop):
        """
        Base class that all capabilities must inherit from.

        :param options: a dictionary of configuration options.

        """
        self.loop = loop
        self.options = options
        self.sessions = {}
        self.users = {}

        self.port = int(options['port'])
        if 'timeout' in options:
            self.timeout = options['timeout']
        else:
            self.timeout = 30

    def create_session(self, address, dest_address):
        protocol = self.__class__.__name__.lower()
        session = Session(address[0], address[1], protocol, self.users, dest_address[1], dest_address[0])
        self.sessions[session.id] = session
        HandlerBase.global_sessions += 1
        logger.debug(
            'Accepted %s session on %s:%s from %s:%s. (%s)',
                protocol, dest_address[0], dest_address[1], address[0],
            address[1], str(session.id))
        logger.debug('Size of session list for %s: %s',
                     protocol, len(self.sessions))
        return session

    def close_session(self, session):
        logger.debug('Closing session. (%s)', str(session.id))
        session.end_session()
        if session.id in self.sessions:
            del self.sessions[session.id]
        else:
            assert False
        HandlerBase.global_sessions -= 1

    async def execute_capability(self, reader, writer, session):
        reader = None  # NOQA
        writer = None  # NOQA
        session = None  # NOQA
        raise Exception("Must be implemented by child")

    async def handle_session(self, reader, writer):
        address = writer.get_extra_info('peername')
        dest_address = writer.get_extra_info('sockname')

        if HandlerBase.global_sessions > HandlerBase.MAX_GLOBAL_SESSIONS:
            protocol = self.__class__.__name__.lower()
            logger.warning(
                'Got %s session on port %s from %s:%s, but not handling it because the global session limit has '
                'been reached', protocol, self.port, *address)
        else:
            session = self.create_session(address, dest_address)
            try:
                await asyncio.wait_for(self.execute_capability(reader, writer, session),
                                       timeout=self.timeout, loop=self.loop)
            except asyncio.TimeoutError:
                logger.debug('Session timed out. (%s)', session.id)
            except (BrokenPipeError, ConnectionError) as err:
                logger.debug('Unexpected end of session: %s, errno: %s. (%s)',
                             err, err.errno, session.id)
            except (UnicodeDecodeError, KeyboardInterrupt):
                pass
            finally:
                self.close_session(session)
                if not self.loop.is_closed():
                    writer.close()



import asyncio
import unittest

import psycopg2
from heralding.capabilities import postgresql
from heralding.misc.common import cancel_all_pending_tasks
from heralding.reporting.reporting_relay import ReportingRelay


class PostgreSQLTests(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

        self.reporting_relay = ReportingRelay()
        self.reporting_relay_task = self.loop.run_in_executor(None, self.reporting_relay.start)

    def tearDown(self):
        self.reporting_relay.stop()
        # We give reporting_relay a chance to be finished
        self.loop.run_until_complete(self.reporting_relay_task)

        self.server.close()
        self.loop.run_until_complete(self.server.wait_closed())

        self.loop.run_until_complete(cancel_all_pending_tasks(self.loop))
        self.loop.close()

    def test_invalid_login(self):
        """Tests if postgres server responds correctly to a invalid login attempt."""

        def postgresql_login():
            try:
                psycopg2.connect("postgres://scott:tiger@0.0.0.0:2504/")
            except psycopg2.OperationalError as e:
                return e
            return None

        options = {'enabled': 'True', 'port': 2504}
        postgresql_cap = postgresql.PostgreSQL(options, self.loop)

        server_coro = asyncio.start_server(
            postgresql_cap.handle_session, '0.0.0.0', 2504, loop=self.loop)
        self.server = self.loop.run_until_complete(server_coro)

        postgresql_task = self.loop.run_in_executor(None, postgresql_login)
        login_exception = self.loop.run_until_complete(postgresql_task)

        self.assertIsInstance(login_exception, psycopg2.OperationalError)
        self.assertEqual(
            str(login_exception),
            'FATAL:  password authentication failed for user "scott"\n'
        )

    def cb(self, socket, command, option):
        return



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import unittest

import telnetlib

from heralding.capabilities import telnet
from heralding.misc.common import cancel_all_pending_tasks
from heralding.reporting.reporting_relay import ReportingRelay


class TelnetTests(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

        self.reporting_relay = ReportingRelay()
        self.reporting_relay_task = self.loop.run_in_executor(None, self.reporting_relay.start)

    def tearDown(self):
        self.reporting_relay.stop()
        # We give reporting_relay a chance to be finished
        self.loop.run_until_complete(self.reporting_relay_task)

        self.server.close()
        self.loop.run_until_complete(self.server.wait_closed())

        self.loop.run_until_complete(cancel_all_pending_tasks(self.loop))
        self.loop.close()

    def test_invalid_login(self):
        """Tests if telnet server responds correctly to a invalid login attempt."""

        def telnet_login():
            client = telnetlib.Telnet('localhost', 2503)
            # set this to 1 if having problems with this test
            client.set_debuglevel(0)
            # this disables all command negotiation.
            client.set_option_negotiation_callback(self.cb)
            # Expect username as first output

            reply = client.read_until(b'Username: ', 1)
            self.assertEqual(b'Username: ', reply)

            client.write(b'someuser' + b'\r\n')
            reply = client.read_until(b'Password: ', 5)
            self.assertTrue(reply.endswith(b'Password: '))

            client.write(b'somepass' + b'\r\n')
            reply = client.read_until(b'\n', 5)
            self.assertTrue(b'\n' in reply)

            client.close()

        options = {'enabled': 'True', 'port': 2503, 'protocol_specific_data': {'max_attempts': 3},
                   'users': {'test': 'test'}}
        telnet_cap = telnet.Telnet(options, self.loop)

        server_coro = asyncio.start_server(telnet_cap.handle_session, '0.0.0.0', 2503, loop=self.loop)
        self.server = self.loop.run_until_complete(server_coro)

        telnet_task = self.loop.run_in_executor(None, telnet_login)
        self.loop.run_until_complete(telnet_task)

    def cb(self, socket, command, option):
        return



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import uuid
import logging
import pprint
from datetime import datetime

import heralding.honeypot
from heralding.reporting.reporting_relay import ReportingRelay

logger = logging.getLogger(__name__)


class Session:
    def __init__(self, source_ip, source_port, protocol, users, destination_port=None, destination_ip=''):

        self.id = uuid.uuid4()
        self.source_ip = source_ip
        self.source_port = source_port
        self.protocol = protocol
        if heralding.honeypot.Honeypot.public_ip:
            self.destination_ip = heralding.honeypot.Honeypot.public_ip
        else:
            self.destination_ip = destination_ip
        self.destination_port = destination_port
        self.timestamp = datetime.utcnow()
        self.num_ = 0
        self.session_ended = False
        # protocol specific data
        self.auxiliary_data = {}

        self.connected = True

        # for session specific volatile data (will not get logged)
        self.vdata = {}
        
        self.auth_attempts = []

        self.last_activity = datetime.utcnow()
        self.log_start_session()

    def log_start_session(self):
        entry = self.get_session_info(False)
        ReportingRelay.logSessionInfo(entry)

    def activity(self):
        self.last_activity = datetime.utcnow()

    def is_connected(self):
        return self.connected

    def get_auxiliary_data(self):
        return {}

    def get_number_of_login_attempts(self):
        return len(self.auth_attempts)

    def add_auth_attempt(self, _type, **kwargs):
        
        # constructs dict to transmitted right away.
        entry = {'timestamp': datetime.utcnow(),
                 'session_id': str(self.id),
                 'auth_id': str(uuid.uuid4()),
                 'source_ip': self.source_ip,
                 'source_port': self.source_port,
                 'destination_ip': self.destination_ip,
                 'destination_port': self.destination_port,
                 'protocol': self.protocol,
                 'username': None,
                 'password': None
                 }
        if 'username' in kwargs:
            entry['username'] = kwargs['username']
        if 'password' in kwargs:
            entry['password'] = kwargs['password']
        ReportingRelay.logAuthAttempt(entry)

        # add to internal dict used for reporting when the session ends
        self.auth_attempts.append({
            'timestamp': entry['timestamp'].strftime('%Y-%m-%d %H:%M:%S.%f'),
            'username': entry['username'],
            'password': entry['password'],
        })

        self.activity()
        logger.debug('%s authentication attempt from %s:%s. Auth mechanism: %s, session id %s '
                     'Credentials: %s', self.protocol, self.source_ip,
                     self.source_port, _type, self.id, json.dumps(kwargs))

    def get_session_info(self, session_ended):
        entry = {'timestamp': self.timestamp.strftime('%Y-%m-%d %H:%M:%S.%f'),
                 'duration': int((datetime.utcnow() - self.timestamp).total_seconds()),
                 'session_id': str(self.id),
                 'source_ip': self.source_ip,
                 'source_port': self.source_port,
                 'destination_ip': self.destination_ip,
                 'destination_port': self.destination_port,
                 'protocol': self.protocol,
                 'num_auth_attempts': len(self.auth_attempts),
                 'auth_attempts': self.auth_attempts,
                 'session_ended': session_ended,
                 'auxiliary_data' : self.auxiliary_data
                 }
        return entry

    def set_auxiliary_data(self, data):
        self.auxiliary_data = data 

    def end_session(self):
        if not self.session_ended:
            self.session_ended = True
            self.connected = False
            entry = self.get_session_info(True)

            ReportingRelay.logSessionInfo(entry)
            logger.debug('Session with session id %s ended', self.id)




# Copyright (C) 2019 Sudipta Pandit <realsdx@protonmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import struct
import logging

from heralding.capabilities.handlerbase import HandlerBase

logger = logging.getLogger(__name__)


# Implemented according to https://dev.mysql.com/doc/internals/en/connection-phase-packets.html
class MySQL(HandlerBase):
    def __init__(self, options, loop):
        super().__init__(options, loop)
        self.PROTO_VER = b'\x0a'
        self.SERVER_VER = b'5.7.16\x00'
        self.AUTH_PLUGIN = b'mysql_native_password\x00'  # SHA-1 hash of password

    @staticmethod
    def convert4To3Byte(num):
        # MySQL protocol requires 3 byte integers for payload length
        return struct.pack("<I", num)[:3]

    def server_greeting(self):
        # Server Greeting(HandshakeV10) Packet:

        # 0x00 0x03    Payload length
        # 0x03 0x01    Sequence No
        # 0x04 0x01    Protocol version (0x0A)
        # 0x05 0x07    MySQL version (5.7.16\0 in this case, variable length in general)
        # 0x0C 0x04    Thread ID
        # 0x10 0x09    Salt (8 random bytes + \0)
        # 0x19 0x02    Server capabilities (0xFF 0x7F)
        # 0x1B 0x01    Server language (0x21 = utf8 COLLATE utf8_general_ci)
        # 0x1C 0x02    Server status (0x02 0x00)
        # 0x1E 0x02    Extended server capabilities (0xFF 0x81)
        # 0x20 0x01    Authentication plugin length
        # 0x21 0x0A    Unused (zeros)
        # 0x2B 0x0D    Salt (12 random bytes + \0)
        # 0x38 0x16    Authentication plugin ('mysql_native_password\0')

        # NB: actual payload without the first four bytes (Payload lenght + sequence no)

        # packet_length including payload_length+sequence_no
        packet_length = 0x3+0x1+0x1+len(self.SERVER_VER)+0x04+(0x08+0x01)+0x02+0x01 + \
            0x02+0x02+0x01+0x0A+0x0D+len(self.AUTH_PLUGIN)

        payload_len = MySQL.convert4To3Byte(packet_length - 0x04)
        seq_no = b'\x00'  # Always will be first packet
        thread_id = struct.pack("<I", 4321)
        salt_1 = os.urandom(8)+b'\x00'
        server_cap = b'\xFF\xF7'
        server_lang = b'\x21'
        server_status = b'\x02\x00'
        ext_server_cap = b'\xFF\x81'
        auth_plugin_len = bytes([len(self.AUTH_PLUGIN)-1])
        zeros = bytes(0x0A)
        salt_2 = os.urandom(12)+b'\x00'

        packet = payload_len+seq_no+self.PROTO_VER+self.SERVER_VER+thread_id +\
            salt_1+server_cap+server_lang+server_status+ext_server_cap+auth_plugin_len +\
            zeros+salt_2+self.AUTH_PLUGIN

        return packet

    def auth_switch_request(self, seq_no):
        # Auth Switch Request Packet:

        # 0x00 0x03    Payload length
        # 0x03 0x01    Sequence No
        # 0x04 0x01    Auth switch request (0xFE)
        # 0x05 0x16    Auth method name (variable length, NUL-terminated)
        # 0x1B 0x15    Auth method data (20 bytes of random salt + \0)

        # Incase any client using other auth methods, request client to switch to 'mysql_native_password'

        packet_length = 3+1+1+len(self.AUTH_PLUGIN)+20+1
        payload_len = MySQL.convert4To3Byte(packet_length-4)
        seq_no = bytes([seq_no])
        auth_switch_req = b'\xFE'
        auth_data = os.urandom(20)+b'\x00'

        packet = payload_len+seq_no+auth_switch_req+self.AUTH_PLUGIN+auth_data

        return packet

    def auth_failed(self, seq_no, user, server, using_password):
        # Auth Failed Packet(Error 1045):

        # 0x00 0x03    Payload length
        # 0x03 0x01    Sequence #
        # 0x04 0x01    Error packet (0xFF)
        # 0x05 0x02    Error code (1045, 0x0415)
        # 0x07 0x06    SQL State (#28000)
        # 0x0D  ??     Error message

        error_msg = bytes("Access denied for user '{}'@'{}' (using password: {})".format(
            user, server, using_password), 'utf-8')
        full_length = 3+1+1+2+6+len(error_msg)  # taking out null

        payload_len = MySQL.convert4To3Byte(full_length-4)
        seq_no = bytes([seq_no])
        error_packet = b'\xFF'  # Error Packet ID
        error_code = b'\x15\x04'  # Error code 1045 (0x0415)
        sql_state = bytes("#28000", 'utf-8')

        packet = payload_len+seq_no+error_packet+error_code+sql_state+error_msg
        return packet

    async def execute_capability(self, reader, writer, session):
        try:
            await self._handle_session(reader, writer, session)
        except struct.error as exc:
            logger.debug('MySQL connection error: %s', exc)
            session.end_session()

    async def _handle_session(self, reader, writer, session):
        # HandshakeResponse41 Packet:

        # 0x00 0x03    Payload length
        # 0x03 0x01    Sequence No
        # 0x04 0x04    Client capability flags
        # 0x08 0x04    Max packet size
        # 0x0C 0x01    Client character set
        # 0x0D 0x17    Unused(zeros)
        # 0x24 ??      User name (NUL-terminated)
        # XX   0x01    Password length
        # XX+1 ??      Password
        # YY   ??      Schema (NUL-terminated)
        # ZZ   ??      Client authentication plugin (NUL-terminated)
        # RR   ??      Client connection attibutes

        address = writer.get_extra_info('peername')[0]
        writer.write(self.server_greeting())
        await writer.drain()
        data = await reader.read(2048)

        # Handle empty data and quit command
        if not data:
            logger.warning("Got no response from client. Ending the session")
            session.end_session()
            return
        elif data == b'\x01\x00\x00\x00\x01':
            logger.warning("Received COM_QUIT from client. Ending the session")
            session.end_session()
            return

        caps = int.from_bytes(data[0x04:0x08], byteorder='little')

        # Check if client version is above 4.1 , if not terminate session
        if not caps & 0x00000200:
            logger.warning("Client vesrion should be 4.1 or above. Ending the session")
            session.end_session()
            return

        try:
            max_size = int.from_bytes(data[0x08:0x0C], byteorder='little')
            username_end_pos = data.index(b'\x00', 0x24, max_size-0x24)
            username = str(data[0x24:username_end_pos], 'utf-8')
            password_len = data[username_end_pos+1]
            using_password = "YES" if password_len > 0 else "NO"
            plugin_offset = username_end_pos+1+1  # start offset of auth_plugin
            seq_no = 2  # if no auth_switch_request
            password_enc = ''

            if password_len > 0:
                # for logging coverted to printable hex
                password_enc = data[plugin_offset:plugin_offset+password_len].hex()
                plugin_offset = plugin_offset+password_len
        except ValueError:
            logger.warning("Malformed packet received. Ending the session")
            session.end_session()
            return

        # check if schema(db) is present
        if caps & 0x00000008:
            schema_offset = plugin_offset  # start offset of db
            try:
                schema_end_pos = data.index(b'\x00', schema_offset, max_size-schema_offset)
                plugin_offset = schema_end_pos+1
            except ValueError:
                logger.warning("Could not find the schema. Ending the session")
                session.end_session()
                return

        # check if plugin_auth enabled
        if caps & 0x00080000:
            try:
                plugin_auth_offset = plugin_offset
                plugin_auth_pos = data.index(b'\x00', plugin_auth_offset, max_size-plugin_auth_offset)
                plugin_auth = data[plugin_auth_offset:plugin_auth_pos]
            except ValueError:
                logger.warning("Cloud not find the plugin_auth data. Ending the session")
                session.end_session()
                return

            if "mysql_native_password" != str(plugin_auth, 'utf-8'):
                writer.write(self.auth_switch_request(seq_no))
                await writer.drain()
                if(await reader.read(1024)):
                    seq_no = 4

        session.add_auth_attempt('encrypted', username=username, password=password_enc)
        writer.write(self.auth_failed(seq_no, username, address, using_password))
        await writer.drain()



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64
import logging
import binascii

from heralding.capabilities.handlerbase import HandlerBase

logger = logging.getLogger(__name__)

CRLF = '\r\n'


class Imap(HandlerBase):
    def __init__(self, options, loop):
        super().__init__(options, loop)
        self.max_tries = int(self.options['protocol_specific_data']['max_attempts'])
        self.banner = self.options['protocol_specific_data']['banner']

        self.available_commands = ['authenticate', 'capability', 'login', 'logout', 'noop']
        self.available_mechanisms = ['plain']

    async def execute_capability(self, reader, writer, session):
        await self._handle_session(session, reader, writer)

    async def _handle_session(self, session, reader, writer,):
        await self.send_message(writer, self.banner)

        state = "Not Authenticated"
        while state != 'Logout' and session.connected:
            # An exception is raised inside await reader.readline() in case of
            # sudden connection reset.
            raw_msg = await reader.readline()
            if not raw_msg:
                break

            raw_msg_str = str(raw_msg, 'utf-8')

            cmd_msg = raw_msg_str.rstrip().split(' ', 2)
            if len(cmd_msg) == 0:
                continue
            elif len(cmd_msg) == 1:
                await self.send_message(writer, "* BAD invalid command")
                continue
            elif len(cmd_msg) == 2:
                tag = cmd_msg[0]
                cmd = cmd_msg[1]
                args = ''
            else:
                tag = cmd_msg[0]
                cmd = cmd_msg[1]
                args = cmd_msg[2]

            cmd = cmd.lower()
            if cmd not in self.available_commands:
                await self.send_message(writer, tag + " BAD invalid command")
            else:
                func_to_call = getattr(self, 'cmd_{0}'.format(cmd), None)
                if func_to_call:
                    return_value = await func_to_call(session, reader, writer, tag, args)
                    state = return_value
                else:
                    await self.send_message(writer, tag + " BAD invalid command")
        session.end_session()

    async def cmd_authenticate(self, session, reader, writer, tag, args):
        mechanism = args.split()
        if len(mechanism) == 1:
            auth_mechanism = mechanism[0].lower()
        else:
            await self.send_message(writer, tag + ' BAD invalid command')
            return "Not Authenticated"

        if auth_mechanism in self.available_mechanisms:
            # the space after '+' is needed according to RFC
            await self.send_message(writer, '+ ')
            raw_msg = await reader.read(512)

            if auth_mechanism == 'plain':
                success, credentials = self.try_b64decode(raw_msg, session)
                # \x00 is a separator between authorization identity,
                # username and password. Authorization identity isn't used in
                # this auth mechanism, so we must have 2 \x00 symbols.(RFC 4616)
                if success and credentials.count('\x00') == 2:
                    raw_msg_dec = str(base64.b64decode(raw_msg), 'utf-8')
                    _, user, password = raw_msg_dec.split('\x00')
                    session.add_auth_attempt('plaintext', username=user, password=password)
                    await self.send_message(writer, tag + ' NO Authentication failed')
                else:
                    await self.send_message(writer, tag + ' BAD invalid command')
        else:
            await self.send_message(writer, tag + ' BAD invalid command')
        self.stop_if_too_many_attempts(session)
        return 'Not Authenticated'

    async def cmd_capability(self, session, reader, writer, tag, args):
        await self.send_message(writer, '* CAPABILITY IMAP4rev1 AUTH=PLAIN')
        await self.send_message(writer, tag + ' OK CAPABILITY completed')
        return 'Not Authenticated'

    async def cmd_login(self, session, reader, writer, tag, args):
        if args:
            user_cred = args.split(' ', 1)
        else:
            await self.send_message(writer, tag + ' BAD invalid command')
            return 'Not Authenticated'

        # Delete first and last quote,
        # because login and password can be sent as quoted strings
        if len(user_cred) == 1:
            user = self.strip_quotes(user_cred[0])
            password = ''
        else:
            user = self.strip_quotes(user_cred[0])
            password = self.strip_quotes(user_cred[1])

        session.add_auth_attempt('plaintext', username=user, password=password)
        await self.send_message(writer, tag + ' NO Authentication failed')
        self.stop_if_too_many_attempts(session)
        return 'Not Authenticated'

    async def cmd_logout(self, session, reader, writer, tag, args):
        await self.send_message(writer, '* BYE IMAP4rev1 Server logging out')
        await self.send_message(writer, tag + ' OK LOGOUT completed')
        return 'Logout'

    async def cmd_noop(self, session, reader, writer, tag, args):
        await self.send_message(writer, tag + ' OK NOOP completed')
        return 'Not Authenticated'

    def stop_if_too_many_attempts(self, session):
        if self.max_tries < session.get_number_of_login_attempts():
            session.end_session()

    @staticmethod
    async def send_message(writer, msg):
        message_bytes = bytes(msg + CRLF, 'utf-8')
        writer.write(message_bytes)
        await writer.drain()

    @staticmethod
    def try_b64decode(b64_str, session):
        try:
            result = base64.b64decode(b64_str)
            return True, str(result, 'utf-8')
        except binascii.Error:
            logger.warning('Error decoding base64: {0} ({1})'.format(binascii.hexlify(b64_str), session.id))
            return False, ''

    @staticmethod
    def strip_quotes(quoted_str):
        if quoted_str.startswith('\"') and quoted_str.endswith('\"'):
            nonquoted_str = quoted_str[1:-1]
        else:
            nonquoted_str = quoted_str
        return nonquoted_str



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import unittest

from heralding.capabilities.pop3 import Pop3
from heralding.misc.common import cancel_all_pending_tasks
from heralding.reporting.reporting_relay import ReportingRelay


class Pop3Tests(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

        self.reporting_relay = ReportingRelay()
        self.reporting_relay_task = self.loop.run_in_executor(None, self.reporting_relay.start)

    def tearDown(self):
        self.reporting_relay.stop()
        # We give reporting_relay a chance to be finished
        self.loop.run_until_complete(self.reporting_relay_task)

        self.server.close()
        self.loop.run_until_complete(self.server.wait_closed())

        self.loop.run_until_complete(cancel_all_pending_tasks(self.loop))

        self.loop.close()

    def test_login(self):
        """Testing different login combinations"""
        async def pop3_login():
            login_sequences = [
                # invalid login, invalid password
                (('USER wakkwakk', b'+OK User accepted'), ('PASS wakkwakk', b'-ERR Authentication failed.')),
                # PASS without user
                (('PASS bond', b'-ERR No username given.'),),
                # Try to run a TRANSACITON state command in AUTHORIZATION state
                (('RETR', b'-ERR Unknown command'),),
            ]
            for sequence in login_sequences:
                reader, writer = await asyncio.open_connection('127.0.0.1', 8888,
                                                               loop=self.loop)
                # skip banner
                await reader.readline()

                for pair in sequence:
                    writer.write(bytes(pair[0] + "\r\n", 'utf-8'))
                    response = await reader.readline()
                    self.assertEqual(response.rstrip(), pair[1])

        options = {'port': 110, 'protocol_specific_data': {'max_attempts': 3}, 'users': {'james': 'bond'}}
        sut = Pop3(options, self.loop)

        server_coro = asyncio.start_server(sut.handle_session, '0.0.0.0', 8888, loop=self.loop)
        self.server = self.loop.run_until_complete(server_coro)

        self.loop.run_until_complete(pop3_login())



# pylint: disable-msg=E1101
# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import curses
import logging


from heralding.capabilities.handlerbase import HandlerBase
from heralding.libs.telnetsrv.telnetsrvlib import TelnetHandlerBase

logger = logging.getLogger(__name__)


class Telnet(HandlerBase):
    def __init__(self, options, loop):
        super().__init__(options, loop)
        TelnetWrapper.max_tries = int(self.options['protocol_specific_data']['max_attempts'])

    async def execute_capability(self, reader, writer, session):
        telnet_cap = TelnetWrapper(reader, writer, session, self.loop)
        await telnet_cap.run()


class TelnetWrapper(TelnetHandlerBase):
    """
    Wraps the telnetsrv module to fit the Honeypot architecture.
    """
    PROMPT = b'$ '
    max_tries = 3
    TERM = 'ansi'

    authNeedUser = True
    authNeedPass = True

    def __init__(self, reader, writer, session, loop):
        self.auth_count = 0
        self.username = None
        self.session = session
        address = writer.get_extra_info('address')
        super().__init__(reader, writer, address, loop)

    async def authentication_ok(self):
        while self.auth_count < TelnetWrapper.max_tries:
            username = await self.readline(prompt=b"Username: ", use_history=False)
            password = await self.readline(echo=False, prompt=b"Password: ", use_history=False)
            self.session.add_auth_attempt(_type='plaintext', username=str(username, 'utf-8'),
                                          password=str(password, 'utf-8'))
            if self.DOECHO:
                self.write(b"\n")
            self.auth_count += 1
        self.writeline(b'Username: ')  # It fixes a problem with Hydra bruteforcer.
        return False

    def setterm(self, term):
        # Dummy file for the purpose of tests.
        with open('/dev/null', 'w') as f:
            curses.setupterm(term, f.fileno())  # This will raise if the termtype is not supported
            self.TERM = term
            self.ESCSEQ = {}
            for k in self.KEYS.keys():
                str_ = curses.tigetstr(curses.has_key._capability_names[k])
                if str_:
                    self.ESCSEQ[str_] = k
            self.CODES['DEOL'] = curses.tigetstr('el')
            self.CODES['DEL'] = curses.tigetstr('dch1')
            self.CODES['INS'] = curses.tigetstr('ich1')
            self.CODES['CSRLEFT'] = curses.tigetstr('cub1')
            self.CODES['CSRRIGHT'] = curses.tigetstr('cuf1')

    def session_end(self):
        self.session.end_session()



# -*- coding: utf-8 -*-
# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest

from heralding.reporting.file_logger import FileLogger


class EncodingTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.auth_log_filename = "test_auth_log.log"
        cls.session_log_filename = "test_session_log.log"
        cls.session_json_filename = "test_session_log.json"
        cls.flogger = FileLogger(cls.session_log_filename, cls.session_json_filename, cls.auth_log_filename)

    def test_ascii(self):
        test_login = test_password = "girls_like_python".encode('ascii')
        test_data = {'username': test_login, 'password': test_password}
        self.flogger.handle_auth_log(test_data)

    def test_unicode(self):
        # word 'python' in russian spelling
        test_login = test_password = "пайтон"
        test_data = {'username': test_login, 'password': test_password}
        self.flogger.handle_auth_log(test_data)

    def test_invalid_utf(self):
        test_login = test_password = "пайт\x80он"
        test_data = {'username': test_login, 'password': test_password}
        self.flogger.handle_auth_log(test_data)

    @classmethod
    def tearDownClass(cls):
        cls.flogger.auth_log_filehandler.close()
        os.remove(cls.auth_log_filename)
        os.remove(cls.session_log_filename)
        os.remove(cls.session_json_filename)



# Copyright (C) 2013 Aniket Panse <contact@aniketpanse.in>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Aniket Panse <contact@aniketpanse.in> grants Johnny Vestergaard <jkv@unixcluster.dk>
# a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable
# copyright license to reproduce, prepare derivative works of, publicly
# display, publicly perform, sublicense, relicense, and distribute [the] Contributions
# and such derivative works.


import base64
import logging

from heralding.capabilities.handlerbase import HandlerBase
from heralding.libs.http.aioserver import AsyncBaseHTTPRequestHandler

logger = logging.getLogger(__name__)


class HTTPHandler(AsyncBaseHTTPRequestHandler):
    def __init__(self, reader, writer, httpsession, options):
        self._options = options
        if 'banner' in self._options:
            self._banner = self._options['banner']
        else:
            self._banner = 'Microsoft-IIS/5.0'
        self._session = httpsession
        address = writer.get_extra_info('address')
        super().__init__(reader, writer, address)

    def do_HEAD(self):
        self.send_response(200)
        self.send_header('Content-type', 'text/html')
        self.end_headers()

    def do_AUTHHEAD(self):
        self.send_response(401)
        # TODO: Value for basic realm...
        self.send_header('WWW-Authenticate', 'Basic realm=\"\"')
        self.send_header('Content-type', 'text/html')
        self.end_headers()

    def do_GET(self):
        if self.headers['Authorization'] is None:
            self.do_AUTHHEAD()
        else:
            hdr = self.headers['Authorization']
            _, enc_uname_pwd = hdr.split(' ')
            dec_uname_pwd = str(base64.b64decode(enc_uname_pwd), 'utf-8')
            uname, pwd = dec_uname_pwd.split(':')
            self._session.add_auth_attempt('plaintext', username=uname, password=pwd)
            self.do_AUTHHEAD()
            headers_bytes = bytes(self.headers['Authorization'], 'utf-8')
            self.wfile.write(headers_bytes)
            self.wfile.write(b'not authenticated')
            aux_data = self.get_auxiliary_info()
            self._session.set_auxiliary_data(aux_data)
            # Disable logging provided by BaseHTTPServer
    def log_message(self, format_, *args):
        pass

    def get_auxiliary_info(self):
        data = {}
        for field in self.headers.keys():
            data.update({str(field) : str(self.headers[str(field)])})

        return data

class Http(HandlerBase):
    def __init__(self, options, loop):
        super().__init__(options, loop)
        self._options = options

    async def execute_capability(self, reader, writer, session):
        http_cap = HTTPHandler(reader, writer, httpsession=session, options=self._options)
        await http_cap.run()
        session.end_session()



# Copyright (C) 2018 Roman Samoilenko <ttahabatt@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from heralding.capabilities.handlerbase import HandlerBase

# Socks constants
SOCKS_VERSION = b'\x05'
AUTH_METHOD = b'\x02'  # username/password authentication. RFC 1929
SOCKS_FAIL = b'\xFF'

logger = logging.getLogger(__name__)


class Socks5(HandlerBase):
    async def execute_capability(self, reader, writer, session):
        await self._handle_session(reader, writer, session)

    async def _handle_session(self, reader, writer, session):
        # 257 - max bytes number for greeting according to RFC 1928
        greeting = await reader.read(257)
        if len(greeting) > 2:
            await self.try_authenticate(reader, writer, session, greeting)
        else:
            logger.debug("Incorrect client greeting string: %r" % greeting)
        session.end_session()

    async def try_authenticate(self, reader, writer, session, greeting):
        version, authmethods = self.unpack_msg(greeting)
        if version == SOCKS_VERSION:
            if AUTH_METHOD in authmethods:
                await self.do_authenticate(reader, writer, session)
            else:
                writer.write(SOCKS_VERSION + SOCKS_FAIL)
                await writer.drain()
            session.set_auxiliary_data(self.get_auxiliary_data(authmethods))
        else:
            logger.debug("Wrong socks version: %r" % version)

    async def do_authenticate(self, reader, writer, session):
        writer.write(SOCKS_VERSION + AUTH_METHOD)
        await writer.drain()
        # 513 - max bytes number for username/password auth according to RFC 1929
        auth_data = await reader.read(513)
        if len(auth_data) > 2:
            username, password = self.unpack_auth(auth_data)
            session.add_auth_attempt('plaintext', username=username.decode(),
                                     password=password.decode())
            writer.write(AUTH_METHOD + SOCKS_FAIL)
            await writer.drain()
        else:
            logger.debug("Wrong authentication data: %r" % auth_data)

    def get_auxiliary_data(self, authmethods):
        _methods = []
        for m in authmethods:
            if m == 2:
                _methods.append("USERNAME/PASSWORD")
            elif m == 0:
                _methods.append("NO AUTHENTICATION REQUIRED")
            elif m == 1:
                _methods.append("GSSAPI")
            elif 3 <= m <= 127:
                _methods.append("IANA ASSIGNED(%s)" % hex(m))
            elif 128 <= m <= 254:
                _methods.append("PRIVATE METHOD(%s)" % hex(m))
            elif m == 255:
                _methods.append("NO ACCEPTABLE METHODS")

        data = {"client_auth_methods": _methods}
        return data

    @staticmethod
    def unpack_msg(data):
        socks_version = data[:1]  # we need byte representation
        authmethods_n = data[1]
        authmethods = data[2:2+authmethods_n]
        return socks_version, authmethods

    @staticmethod
    def unpack_auth(auth_data):
        ulen = auth_data[1]
        username = auth_data[2:2+ulen]
        password_part = auth_data[2+ulen:]
        if password_part:
            plen = password_part[0]
            password = auth_data[-plen:]
        else:
            password = b""
        return username, password



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import hpfeeds
import json

from heralding.reporting.base_logger import BaseLogger

logger = logging.getLogger(__name__)


class HpFeedsLogger(BaseLogger):
    def __init__(self, session_channel, auth_channel, host, port, ident, secret):
        super().__init__()
        self.session_channel = session_channel
        self.auth_channel = auth_channel
        self.host = host
        self.port = port
        self.ident = ident
        self.secret = secret
        self._initial_connection_happend = False
        logger.info('HpFeeds logger started.')

    def start(self):
        if not self._initial_connection_happend:
            self.hp_connection = hpfeeds.new(self.host, self.port, self.ident, self.secret, True)
            self._initial_connection_happend = True
            logger.info('HpFeeds logger connected to %s:%s.', self.host, self.port)
        # after we established that we can connect enter the subscribe and enter the polling loop
        super().start()

    def loggerStopped(self):
        self.stop()
        self.close()

    def handle_auth_log(self, data):
        if self._initial_connection_happend:
            self.hp_connection.publish(self.auth_channel, json.dumps(data).encode())

    def handle_session_log(self, data):
        if self._initial_connection_happend:
            self.hp_connection.publish(self.session_channel, json.dumps(data).encode())



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from heralding.capabilities.handlerbase import HandlerBase

logger = logging.getLogger(__name__)


class Pop3(HandlerBase):
    max_tries = 10
    cmds = {}

    def __init__(self, options, loop):
        super().__init__(options, loop)
        Pop3.max_tries = int(self.options['protocol_specific_data']['max_attempts'])

    async def execute_capability(self, reader, writer, session):
        await self._handle_session(session, reader, writer)

    async def _handle_session(self, session, reader, writer):
        await self.send_message(writer, '+OK POP3 server ready')

        state = 'AUTHORIZATION'
        while state != '' and session.connected:
            raw_msg = await reader.readline()
            if not raw_msg:
                break

            raw_msg_str = str(raw_msg, 'utf-8')

            session.activity()
            cmd_msg = raw_msg_str.rstrip().split(' ', 1)
            if len(cmd_msg) == 0:
                continue
            elif len(cmd_msg) == 1:
                cmd = cmd_msg[0]
                msg = ''
            else:
                cmd = cmd_msg[0]
                msg = cmd_msg[1]

            cmd = cmd.lower()

            if cmd not in ['user', 'pass', 'quit', 'noop']:
                await self.send_message(writer, '-ERR Unknown command')
            else:
                func_to_call = getattr(self, 'cmd_{0}'.format(cmd), None)
                return_value = await func_to_call(session, reader, writer, msg)
                # state changers!
                if state == 'AUTHORIZATION' or cmd == 'quit':
                    state = return_value

        session.end_session()

    # APOP mrose c4c9334bac560ecc979e58001b3e22fb
    # +OK mrose's maildrop has 2 messages (320 octets)
    def auth_apop(self, session, gsocket, msg):
        raise Exception('Not implemented yet!')

    # USER mrose
    # +OK User accepted
    # PASS tanstaaf
    # +OK Pass accepted
    # or: "-ERR Authentication failed."
    # or: "-ERR No username given."
    async def cmd_user(self, session, reader, writer, msg):
        session.vdata['USER'] = msg
        await self.send_message(writer, '+OK User accepted')
        return 'AUTHORIZATION'

    async def cmd_pass(self, session, reader, writer, msg):
        if 'USER' not in session.vdata:
            await self.send_message(writer, '-ERR No username given.')
        else:
            session.add_auth_attempt('plaintext', username=session.vdata['USER'], password=msg)
            await self.send_message(writer, "-ERR Authentication failed.")

        if 'USER' in session.vdata:
            del session.vdata['USER']
        return 'AUTHORIZATION'

    async def cmd_noop(self, session, reader, writer, msg):
        await self.send_message(writer, '+OK')

    async def cmd_quit(self, session, reader, writer, msg):
        await self.send_message(writer, '+OK Logging out')
        return ''

    @staticmethod
    async def send_message(writer, msg):
        message_bytes = bytes(msg + "\n", 'utf-8')
        writer.write(message_bytes)
        await writer.drain()



# Copyright (C) 2018 Roman Samoilenko <ttahabatt@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import unittest

import heralding.capabilities.socks5 as socks
from heralding.reporting.reporting_relay import ReportingRelay


class Socks5Tests(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

        self.reporting_relay = ReportingRelay()
        self.reporting_relay_task = self.loop.run_in_executor(None, self.reporting_relay.start)

    def tearDown(self):
        self.reporting_relay.stop()
        # We give reporting_relay a chance to be finished
        self.loop.run_until_complete(self.reporting_relay_task)

        self.server.close()
        self.loop.run_until_complete(self.server.wait_closed())

        self.loop.close()

    def test_socks_authentication(self):
        async def socks_auth():
            reader, writer = await asyncio.open_connection('127.0.0.1', 8888,
                                                           loop=self.loop)

            # Greeting to the server. version+authmethod number+authmethod
            client_greeting = socks.SOCKS_VERSION + b"\x01" + socks.AUTH_METHOD
            writer.write(client_greeting)

            # Receive version+chosen authmethod
            _ = await reader.read(2)

            # Send credentials.
            # version+username len+username+password len+password
            credentials = b"\x05\x08username\x08password"
            writer.write(credentials)

            # Receive authmethod+\xff
            res = await reader.read(2)
            self.assertEqual(res, socks.AUTH_METHOD + socks.SOCKS_FAIL)

        options = {'enabled': 'True', 'port': 8888, 'timeout': 30}
        capability = socks.Socks5(options, self.loop)

        server_coro = asyncio.start_server(capability.handle_session, '127.0.0.1',
                                           8888, loop=self.loop)
        self.server = self.loop.run_until_complete(server_coro)
        self.loop.run_until_complete(socks_auth())



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import csv
import logging
import json

from heralding.reporting.base_logger import BaseLogger
import heralding

logger = logging.getLogger(__name__)


class FileLogger(BaseLogger):
    def __init__(self, session_csv_logfile, sessions_json_logfile, auth_logfile):
        super().__init__()

        self.auth_log_filehandler = None
        self.auth_log_writer = None
        self.session_csv_log_filehandler = None 
        self.session_csv_log_writer = None
        self.session_json_log_filehandler = None

        if auth_logfile != "":
            # Setup CSV logging for auth attempts
            auth_field_names = ['timestamp', 'auth_id', 'session_id', 'source_ip', 'source_port', 'destination_ip',
                                'destination_port', 'protocol', 'username', 'password']
            
            self.auth_log_filehandler, self.auth_log_writer = self.setup_csv_files(
                auth_logfile, auth_field_names)

            logger.info(
                'File logger: Using %s to log authentication attempts in CSV format.', auth_logfile)

        if session_csv_logfile != "":
            # Setup CSV logging for sessions
            session_field_names = ['timestamp', 'duration', 'session_id', 'source_ip', 'source_port', 'destination_ip',
                                'destination_port', 'protocol', 'num_auth_attempts']
            self.session_csv_log_filehandler, self.session_csv_log_writer = self.setup_csv_files(
                session_csv_logfile, session_field_names)

            logger.info('File logger: Using %s to log unified session data in CSV format.',
                        session_csv_logfile)

        if sessions_json_logfile != "":
            # Setup json logging for logging complete sessions
            if not os.path.isfile(sessions_json_logfile):
                self.session_json_log_filehandler = open(sessions_json_logfile, 'w', encoding='utf-8')
            else:
                self.session_json_log_filehandler = open(sessions_json_logfile, 'a', encoding='utf-8')
            
            logger.info('File logger: Using %s to log complete session data in JSON format.',
                        sessions_json_logfile)

    def setup_csv_files(self, filename, field_names):
        handler = writer = None

        if not os.path.isfile(filename):
            handler = open(filename, 'w', encoding='utf-8')
        else:
            handler = open(filename, 'a', encoding='utf-8')

        writer = csv.DictWriter(
            handler, fieldnames=field_names, extrasaction='ignore')

        # empty file, write csv header
        if os.path.getsize(filename) == 0:
            writer.writeheader()
            handler.flush()

        return handler, writer

    def loggerStopped(self):
        for handler in [self.auth_log_filehandler, self.session_csv_log_filehandler, 
                        self.session_json_log_filehandler]:
            if handler != None:
                handler.flush()
                handler.close()

    def handle_auth_log(self, data):
        # for now this logger only handles authentication attempts where we are able
        # to log both username and password
        if self.auth_log_filehandler != None:
            if 'username' in data and 'password' in data:
                self.auth_log_writer.writerow(data)
                # meh
                self.auth_log_filehandler.flush()

    def handle_session_log(self, data):
        if data['session_ended']:
            if self.session_csv_log_filehandler != None:
                self.session_csv_log_writer.writerow(data)
                self.session_csv_log_filehandler.flush()
            if self.session_json_log_filehandler != None:
                self.session_json_log_filehandler.write(json.dumps(data) + "\n")
                self.session_json_log_filehandler.flush()




# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# Rewritten by Aniket Panse <contact@aniketpanse.in>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Aniket Panse <contact@aniketpanse.in> grants Johnny Vestergaard <jkv@unixcluster.dk>
# a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable
# copyright license to reproduce, prepare derivative works of, publicly
# display, publicly perform, sublicense, relicense, and distribute [the] Contributions
# and such derivative works.

import logging

from heralding.capabilities.handlerbase import HandlerBase

logger = logging.getLogger(__name__)

TERMINATOR = '\r\n'


class FtpHandler:
    """Handles a single FTP connection"""

    def __init__(self, reader, writer, options, session):
        self.banner = options['protocol_specific_data']['banner']
        self.max_loggins = int(options['protocol_specific_data']['max_attempts'])
        self.syst_type = options['protocol_specific_data']['syst_type']
        self.authenticated = False
        self.writer = writer
        self.reader = reader
        self.serve_flag = True
        self.session = session

        self.state = None
        self.user = None

    async def getcmd(self):
        cmd = await self.reader.readline()
        return str(cmd, 'utf-8')

    async def serve(self):
        await self.respond('220 ' + self.banner)

        while self.serve_flag:
            resp = await self.getcmd()
            if not resp:
                self.stop()
                break
            else:
                try:
                    cmd, args = resp.split(' ', 1)
                except ValueError:
                    cmd = resp
                    args = None
                else:
                    args = args.strip('\r\n')
                cmd = cmd.strip('\r\n')
                cmd = cmd.upper()
                # List of commands allowed before a login
                unauth_cmds = ['USER', 'PASS', 'QUIT', 'SYST']
                meth = getattr(self, 'do_' + cmd, None)
                if not meth:
                    await self.respond('500 Unknown Command.')
                else:
                    if not self.authenticated:
                        if cmd not in unauth_cmds:
                            await self.respond('503 Login with USER first.')
                            continue
                    await meth(args)
                    self.state = cmd

    async def do_USER(self, arg):
        self.user = arg
        await self.respond('331 Now specify the Password.')

    async def do_PASS(self, arg):
        if self.state != 'USER':
            await self.respond('503 Login with USER first.')
            return
        passwd = arg
        self.session.add_auth_attempt('plaintext', username=self.user, password=passwd)
        await self.respond('530 Authentication Failed.')
        if self.session.get_number_of_login_attempts() >= self.max_loggins:
            self.stop()

    async def do_SYST(self, arg):
        await self.respond('215 {0}'.format(self.syst_type))

    async def do_QUIT(self, arg):
        await self.respond('221 Bye.')
        self.serve_flag = False
        self.stop()

    async def respond(self, msg):
        msg += TERMINATOR
        msg_bytes = bytes(msg, 'utf-8')
        self.writer.write(msg_bytes)
        await self.writer.drain()

    def stop(self):
        self.session.end_session()


class ftp(HandlerBase):
    def __init__(self, options, loop):
        super().__init__(options, loop)
        self._options = options

    async def execute_capability(self, reader, writer, session):
        ftp_cap = FtpHandler(reader, writer, self._options, session)
        await ftp_cap.serve()



# Copyright (C) 2012 Aniket Panse <contact@aniketpanse.in
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Aniket Panse <contact@aniketpanse.in> grants Johnny Vestergaard <jkv@unixcluster.dk>
# a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable
# copyright license to reproduce, prepare derivative works of, publicly
# display, publicly perform, sublicense, relicense, and distribute [the] Contributions
# and such derivative works.

import hmac
import base64
import asyncio
import smtplib
import unittest

from heralding.capabilities import smtp
from heralding.reporting.reporting_relay import ReportingRelay


class SmtpTests(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

        self.reporting_relay = ReportingRelay()
        self.reporting_relay_task = self.loop.run_in_executor(None, self.reporting_relay.start)

    def tearDown(self):
        self.reporting_relay.stop()
        # We give reporting_relay a chance to be finished
        self.loop.run_until_complete(self.reporting_relay_task)

        self.server.close()
        self.loop.run_until_complete(self.server.wait_closed())

        self.loop.close()

    def test_connection(self):
        """ Tries to connect and run a EHLO command. Very basic test.
        """
        def smtp_connection():
            smtp_ = smtplib.SMTP('127.0.0.1', 8888, local_hostname='localhost', timeout=15)
            smtp_.ehlo()
            smtp_.quit()

        # Use uncommon port so that we can run test even if the Honeypot is running.
        options = {'enabled': 'True', 'port': 8888, 'protocol_specific_data': {'banner': 'test'},
                   'users': {'test': 'test'}, }
        smtp_cap = smtp.smtp(options, self.loop)

        server_coro = asyncio.start_server(smtp_cap.handle_session, '0.0.0.0', 8888, loop=self.loop)
        self.server = self.loop.run_until_complete(server_coro)

        smtp_task = self.loop.run_in_executor(None, smtp_connection)
        self.loop.run_until_complete(smtp_task)

    def test_AUTH_CRAM_MD5_reject(self):
        """ Makes sure the server rejects all invalid login attempts that use the
            CRAM-MD5 Authentication method.
        """
        def encode_cram_md5(challenge, user, password):
            challenge = base64.decodebytes(challenge)
            response = user + b' ' + bytes(hmac.HMAC(password, challenge).hexdigest(), 'utf-8')
            return str(base64.b64encode(response), 'utf-8')

        def smtp_auth_cram_md5():
            smtp_ = smtplib.SMTP('127.0.0.1', 8888, local_hostname='localhost', timeout=15)
            _, resp = smtp_.docmd('AUTH', 'CRAM-MD5')
            code, resp = smtp_.docmd(encode_cram_md5(resp, b'test', b'test'))
            smtp_.quit()
            # For now, the server's going to return a 535 code.
            self.assertEqual(code, 535)

        options = {'enabled': 'True', 'port': 8888, 'protocol_specific_data': {'banner': 'Test'},
                   'users': {'someguy': 'test'}}
        smtp_cap = smtp.smtp(options, self.loop)

        server_coro = asyncio.start_server(smtp_cap.handle_session, '0.0.0.0', 8888, loop=self.loop)
        self.server = self.loop.run_until_complete(server_coro)

        smtp_task = self.loop.run_in_executor(None, smtp_auth_cram_md5)
        self.loop.run_until_complete(smtp_task)

    def test_AUTH_PLAIN_reject(self):
        """ Makes sure the server rejects all invalid login attempts that use the PLAIN Authentication method.
        """
        def smtp_auth_plain_reject():
            smtp_ = smtplib.SMTP('127.0.0.1', 8888, local_hostname='localhost', timeout=15)
            arg = bytes('\0{0}\0{1}'.format('test', 'test'), 'utf-8')
            code, _ = smtp_.docmd('AUTH', 'PLAIN ' + str(base64.b64encode(arg), 'utf-8'))
            smtp_.quit()
            self.assertEqual(code, 535)

        options = {'enabled': 'True', 'port': 0, 'protocol_specific_data': {'banner': 'Test'},
                   'users': {'someguy': 'test'}}

        smtp_cap = smtp.smtp(options, self.loop)

        server_coro = asyncio.start_server(smtp_cap.handle_session, '0.0.0.0', 8888, loop=self.loop)
        self.server = self.loop.run_until_complete(server_coro)

        smtp_task = self.loop.run_in_executor(None, smtp_auth_plain_reject)
        self.loop.run_until_complete(smtp_task)

    def test_AUTH_LOGIN_reject(self):
        """ Makes sure the server rejects all invalid login attempts that use the LOGIN Authentication method.
        """
        def smtp_auth_login_reject():
            smtp_ = smtplib.SMTP('127.0.0.1', 8888, local_hostname='localhost', timeout=15)
            smtp_.docmd('AUTH', 'LOGIN')
            smtp_.docmd(str(base64.b64encode(b'test'), 'utf-8'))
            code, _ = smtp_.docmd(str(base64.b64encode(b'test'), 'utf-8'))
            smtp_.quit()
            self.assertEqual(code, 535)

        options = {'enabled': 'True', 'port': 0, 'protocol_specific_data': {'banner': 'Test'},
                   'users': {'someguy': 'test'}}

        smtp_cap = smtp.smtp(options, self.loop)

        server_coro = asyncio.start_server(smtp_cap.handle_session, '0.0.0.0', 8888, loop=self.loop)
        self.server = self.loop.run_until_complete(server_coro)

        smtp_task = self.loop.run_in_executor(None, smtp_auth_login_reject)
        self.loop.run_until_complete(smtp_task)


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from heralding.reporting.hpfeeds_logger import HpFeedsLogger


class FtpTests(unittest.TestCase):
    def test_hpfeeds(self):
        """Basic test for hpfeeds reporter"""

        session_channel = "heralding.session"
        auth_channel = "heraldign.auth"
        host = "127.0.0.1"
        port = 12345
        ident = "atzqøl"
        secret = "toosecret"

        HpFeedsLogger(session_channel, auth_channel, host, port, ident, secret)



# Copyright (C) 2018 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import heralding.misc

import logging
import zmq
import json
from datetime import datetime

from heralding.reporting.base_logger import BaseLogger

logger = logging.getLogger(__name__)


class CuriosumIntegration(BaseLogger):
    def __init__(self, port):
        super().__init__()

        zmq_socket = 'tcp://127.0.0.1:{0}'.format(port)

        context = heralding.misc.zmq_context
        self.socket = context.socket(zmq.PUSH)
        self.socket.bind(zmq_socket)
        self.listen_ports = []
        self.last_listen_ports_transmit = datetime.now()

        logger.info(
            'Curiosum logger started using files: %s', zmq_socket)

    def loggerStopped(self):
        self.socket.close()

    def _no_block_send(self, topic, data):
        try:
            self.socket.send_string('{0} {1}'.format(
                topic, json.dumps(data)), zmq.NOBLOCK)
        except zmq.ZMQError as e:
            logger.warning('Error while sending: %s', e)

    def handle_session_log(self, data):
        message = {
            'SessionID': str(data['session_id']),
            'DstPort': data['destination_port'],
            'SrcIP': data['source_ip'],
            'SrcPort': data['source_port'],
            'SessionEnded': data['session_ended']}
        self._no_block_send('session_ended', message)

    def _execute_regulary(self):
        if (datetime.now() - self.last_listen_ports_transmit).total_seconds() > 5:
            self._no_block_send('listen_ports', self.listen_ports)
            self.last_listen_ports_transmit = datetime.now()

    def handle_listen_ports(self, data):
        self.listen_ports = data



import struct
import logging

from heralding.capabilities.handlerbase import HandlerBase

logger = logging.getLogger(__name__)


class PostgreSQL(HandlerBase):
    async def execute_capability(self, reader, writer, session):
        try:
            await self._handle_session(session, reader, writer)
        except struct.error as exc:
            logger.debug('PostgreSQL connection error: %s', exc)
            session.end_session()

    async def _handle_session(self, session, reader, writer):
        # Read (we blindly assume) SSL request and deny it
        await read_msg(reader, writer)

        writer.write(b'N')

        session.activity()

        # Read login details
        data = await read_msg(reader, writer)
        login_dict = parse_dict(data)

        # Request plain text password login
        password_request = [b'R', 8, 3]
        writer.write(struct.pack('>c I I', *password_request))
        await writer.drain()

        # Read password
        data = await read_msg(reader, writer)
        password = parse_str(data)
        username = login_dict['user']
        session.add_auth_attempt('plaintext', username=username, password=password)

        # Report login failure
        writer.write(b'E')
        fail = [
            b'SFATAL\x00C28P01\x00',
            'Mpassword authentication failed for user "{}"'.format(username).encode('utf-8'),
            b'\x00Fauth.c\x00L288\x00Rauth_failed\x00\x00',
        ]

        length = sum([len(f) for f in fail])
        writer.write(struct.pack('>I', length + 4))
        for f in fail:
            writer.write(f)

        await writer.drain()

        session.end_session()


async def read_msg(reader, writer):
    i = await reader.read(4)
    length = struct.unpack('>I', i)[0]
    data = await reader.read(length)
    return data


def parse_dict(data):
    dct = {}
    mode = 'pad'
    key = []
    value = []

    for c in struct.iter_unpack('c', data):
        c = c[0]

        if mode == 'pad':
            if c in (bytes([0]), bytes([3])):
                continue
            else:
                mode = 'key'

        if mode == 'key':
            if c == bytes([0]):
                mode = 'value'
            else:
                key.append(c.decode())

        elif mode == 'value':
            if c == bytes([0]):
                dct[''.join(key)] = ''.join(value)
                key = []
                value = []
                mode = 'pad'
            else:
                value.append(c.decode())

    return dct


def parse_str(data):
    data_array = bytearray(data)
    return data_array[1:-1].decode('utf-8')



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import ssl
import sys
import logging
import asyncio

import heralding.misc.common as common
import heralding.capabilities.handlerbase

from heralding.reporting.reporting_relay import ReportingRelay
from heralding.reporting.file_logger import FileLogger
from heralding.reporting.syslog_logger import SyslogLogger
from heralding.reporting.hpfeeds_logger import HpFeedsLogger
from heralding.reporting.curiosum_integration import CuriosumIntegration

import asyncssh

from ipify import get_ip

logger = logging.getLogger(__name__)


class Honeypot:
    public_ip = ''

    def __init__(self, config, loop):
        """
        :param config: configuration dictionary.
        """
        assert config is not None
        self.loop = loop
        self.SshClass = None
        self.config = config
        self._servers = []
        self._loggers = []

    async def _record_and_lookup_public_ip(self):
        while True:
            try:
                Honeypot.public_ip = get_ip()
                logger.warning('Found public ip: %s', Honeypot.public_ip)
            except Exception as ex:
                Honeypot.public_ip = ''
                logger.warning('Could not request public ip from ipify, error: %s', ex)
            await asyncio.sleep(3600)

    def start(self):
        """ Starts services. """

        if 'public_ip_as_destination_ip' in self.config and self.config['public_ip_as_destination_ip'] is True:
            asyncio.ensure_future(self._record_and_lookup_public_ip(), loop=self.loop)

        # start activity logging
        if 'activity_logging' in self.config:
            if 'file' in self.config['activity_logging'] and self.config['activity_logging']['file']['enabled']:
                auth_log = self.config['activity_logging']['file']['authentication_log_file']
                session_csv_log = self.config['activity_logging']['file']['session_csv_log_file']
                session_json_log = self.config['activity_logging']['file']['session_json_log_file']
                file_logger = FileLogger(session_csv_log, session_json_log, auth_log)
                self.file_logger_task = self.loop.run_in_executor(None, file_logger.start)
                self.file_logger_task.add_done_callback(common.on_unhandled_task_exception)
                self._loggers.append(file_logger)

            if 'syslog' in self.config['activity_logging'] and self.config['activity_logging']['syslog']['enabled']:
                sys_logger = SyslogLogger()
                self.sys_logger_task = self.loop.run_in_executor(None, sys_logger.start)
                self.sys_logger_task.add_done_callback(common.on_unhandled_task_exception)
                self._loggers.append(sys_logger)

            if 'hpfeeds' in self.config['activity_logging'] and self.config['activity_logging']['hpfeeds']['enabled']:
                session_channel = self.config['activity_logging']['hpfeeds']['session_channel']
                auth_channel = self.config['activity_logging']['hpfeeds']['auth_channel']
                host = self.config['activity_logging']['hpfeeds']['host']
                port = self.config['activity_logging']['hpfeeds']['port']
                ident = self.config['activity_logging']['hpfeeds']['ident']
                secret = self.config['activity_logging']['hpfeeds']['secret']
                hpfeeds_logger = HpFeedsLogger(session_channel, auth_channel, host, port, ident, secret)
                self.hpfeeds_logger_task = self.loop.run_in_executor(None, hpfeeds_logger.start)
                self.hpfeeds_logger_task.add_done_callback(common.on_unhandled_task_exception)

            if 'curiosum' in self.config['activity_logging'] and self.config['activity_logging']['curiosum']['enabled']:
                port = self.config['activity_logging']['curiosum']['port']
                curiosum_integration = CuriosumIntegration(port)
                self.hpfeeds_logger_task = self.loop.run_in_executor(None, curiosum_integration.start)
                self.hpfeeds_logger_task.add_done_callback(common.on_unhandled_task_exception)

        bind_host = self.config['bind_host']
        listen_ports = []
        for c in heralding.capabilities.handlerbase.HandlerBase.__subclasses__():
            cap_name = c.__name__.lower()
            if cap_name in self.config['capabilities']:
                if not self.config['capabilities'][cap_name]['enabled']:
                    continue
                port = self.config['capabilities'][cap_name]['port']
                listen_ports.append(port)
                # carve out the options for this specific service
                options = self.config['capabilities'][cap_name]
                # capabilities are only allowed to append to the session list
                cap = c(options, self.loop)

                try:
                    # # Convention: All capability names which end in 's' will be wrapped in ssl.
                    if cap_name.endswith('s'):
                        pem_file = '{0}.pem'.format(cap_name)
                        self.create_cert_if_not_exists(cap_name, pem_file)
                        ssl_context = self.create_ssl_context(pem_file)
                        server_coro = asyncio.start_server(cap.handle_session, bind_host, port,
                                                           loop=self.loop, ssl=ssl_context)
                    elif cap_name == 'ssh':
                        # Since dicts and user-defined classes are mutable, we have
                        # to save ssh class and ssh options somewhere.
                        ssh_options = options
                        SshClass = c
                        self.SshClass = SshClass

                        ssh_key_file = 'ssh.key'
                        SshClass.generate_ssh_key(ssh_key_file)

                        banner = ssh_options['protocol_specific_data']['banner']
                        SshClass.change_server_banner(banner)

                        server_coro = asyncssh.create_server(lambda: SshClass(ssh_options, self.loop),
                                                             bind_host, port, server_host_keys=[ssh_key_file],
                                                             login_timeout=cap.timeout, loop=self.loop)
                    else:
                        server_coro = asyncio.start_server(cap.handle_session, bind_host, port, loop=self.loop)

                    server = self.loop.run_until_complete(server_coro)
                    logger.debug('Adding %s capability with options: %s', cap_name, options)
                    self._servers.append(server)
                except Exception as ex:
                    error_message = "Could not start {0} server on port {1}. Error: {2}".format(c.__name__, port, ex)
                    logger.error(error_message)
                    self.loop.run_until_complete(common.cancel_all_pending_tasks(self.loop))
                    sys.exit(error_message)
                else:
                    logger.info('Started %s capability listening on port %s', c.__name__, port)
        ReportingRelay.logListenPorts(listen_ports)

    def stop(self):
        """Stops services"""
        if self.config['capabilities']['ssh']['enabled']:
            for conn in self.SshClass.connections_list:
                conn.close()
                self.loop.run_until_complete(conn.wait_closed())

        for server in self._servers:
            server.close()
            self.loop.run_until_complete(server.wait_closed())

        for l in self._loggers:
            l.stop()

        self.loop.run_until_complete(common.cancel_all_pending_tasks(self.loop))

        logger.info('All tasks were stopped.')

    def create_cert_if_not_exists(self, cap_name, pem_file):
        if not os.path.isfile(pem_file):
            logger.debug('Generating certificate and key: %s', pem_file)

            cert_dict = self.config['capabilities'][cap_name]['protocol_specific_data']['cert']
            cert_cn = cert_dict['common_name']
            cert_country = cert_dict['country']
            cert_state = cert_dict['state']
            cert_locality = cert_dict['locality']
            cert_org = cert_dict['organization']
            cert_org_unit = cert_dict['organizational_unit']
            valid_days = int(cert_dict['valid_days'])
            serial_number = int(cert_dict['serial_number'])

            cert, key = common.generate_self_signed_cert(cert_country, cert_state, cert_org, cert_locality,
                                                         cert_org_unit, cert_cn, valid_days, serial_number)
            with open(pem_file, 'wb') as _pem_file:
                _pem_file.write(cert)
                _pem_file.write(key)

    @staticmethod
    def create_ssl_context(pem_file):
        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_context.check_hostname = False
        ssl_context.load_cert_chain(pem_file)
        return ssl_context



import asyncio
import unittest

import pymysql
from heralding.capabilities import mysql
from heralding.misc.common import cancel_all_pending_tasks
from heralding.reporting.reporting_relay import ReportingRelay


class MySQLTests(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

        self.reporting_relay = ReportingRelay()
        self.reporting_relay_task = self.loop.run_in_executor(None, self.reporting_relay.start)

    def tearDown(self):
        self.reporting_relay.stop()
        # We give reporting_relay a chance to be finished
        self.loop.run_until_complete(self.reporting_relay_task)

        self.server.close()
        self.loop.run_until_complete(self.server.wait_closed())

        self.loop.run_until_complete(cancel_all_pending_tasks(self.loop))
        self.loop.close()

    def test_invalid_login(self):
        """Tests if mysql server responds correctly to a invalid login attempt."""

        def mysql_login():
            try:
                pymysql.connect(host="0.0.0.0",
                                port=8306,
                                user="tuser",
                                password="tpass",
                                db="testdb")
            except pymysql.err.OperationalError as e:
                return e
            return None

        options = {'enabled': 'True', 'port': 8306}
        mysql_cap = mysql.MySQL(options, self.loop)

        server_coro = asyncio.start_server(
            mysql_cap.handle_session, '0.0.0.0', 8306, loop=self.loop)
        self.server = self.loop.run_until_complete(server_coro)

        mysql_task = self.loop.run_in_executor(None, mysql_login)
        login_exception = self.loop.run_until_complete(mysql_task)

        self.assertIsInstance(login_exception, pymysql.err.OperationalError)
        self.assertEqual(
            str(login_exception),
            '(1045, "Access denied for user \'tuser\'@\'127.0.0.1\' (using password: YES)")'
        )

    def cb(self, socket, command, option):
        return



# Copyright (C) 2017 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging
import asyncio

from OpenSSL import crypto
from Crypto.PublicKey import RSA

logger = logging.getLogger(__name__)


def on_unhandled_task_exception(task):
    if not task.cancelled():
        task_exc = task.exception()
        if task_exc:
            logger.error('Stopping because %s died: %s', task, task_exc)
            os._exit(1)


async def cancel_all_pending_tasks(loop=None):
    if loop is None:
        loop = asyncio.get_event_loop()
    pending = asyncio.Task.all_tasks(loop=loop)
    pending.remove(asyncio.Task.current_task(loop=loop))
    for task in pending:
        # We give task only 1 second to die.
        task.cancel()
        try:
            await asyncio.wait_for(task, timeout=1, loop=loop)
        except (asyncio.CancelledError, KeyboardInterrupt, ConnectionResetError):
            pass


def generate_self_signed_cert(cert_country, cert_state, cert_organization, cert_locality, cert_organizational_unit,
                              cert_common_name, valid_days, serial_number):
    rsa_key = RSA.generate(2048)

    pk = crypto.load_privatekey(crypto.FILETYPE_PEM, rsa_key.exportKey('PEM', pkcs=1))
    cert = crypto.X509()
    sub = cert.get_subject()
    sub.CN = cert_common_name
    sub.C = cert_country
    sub.ST = cert_state
    sub.L = cert_locality
    sub.O = cert_organization

    # optional
    if cert_organizational_unit:
        sub.OU = cert_organizational_unit

    cert.set_serial_number(serial_number)
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(valid_days * 24 * 60 * 60)  # Valid for a year
    cert.set_issuer(sub)
    cert.set_pubkey(pk)
    cert.sign(pk, 'sha1')

    cert_text = crypto.dump_certificate(crypto.FILETYPE_PEM, cert)
    priv_key_text = rsa_key.exportKey('PEM', pkcs=1)

    return cert_text, priv_key_text



# license: LGPL
# Thanks a lot to the author of original telnetsrvlib - Ian Epperson (https://github.com/ianepperson)!
# Original repository - https://github.com/ianepperson/telnetsrvlib
# We need this in order to reduce the number of third-party modules.
# This code is adjusted to work with asyncio in our specific case.

import curses
import logging
import asyncio
import curses.ascii
import curses.has_key

from heralding.libs.aiobaserequest import AsyncBaseRequestHandler

log = logging.getLogger(__name__)

BELL = bytes([7])
ANSI_START_SEQ = b'['
ANSI_KEY_TO_CURSES = {
    b'A': curses.KEY_UP,
    b'B': curses.KEY_DOWN,
    b'C': curses.KEY_RIGHT,
    b'D': curses.KEY_LEFT,
}

# Telnet protocol characters (don't change)
IAC = bytes([255])  # "Interpret As Command"
DONT = bytes([254])
DO = bytes([253])
WONT = bytes([252])
WILL = bytes([251])
theNULL = bytes([0])

SE = bytes([240])  # Subnegotiation End
NOP = bytes([241])  # No Operation
SB = bytes([250])  # Subnegotiation Begin

# Telnet protocol options code (don't change)
# These ones all come from arpa/telnet.h
ECHO = bytes([1])  # echo
SGA = bytes([3])  # suppress go ahead
TTYPE = bytes([24])  # terminal type
NAWS = bytes([31])  # window size
LINEMODE = bytes([34])  # Linemode option
NEW_ENVIRON = bytes([39])  # New - Environment variables
NOOPT = bytes([0])

IS = bytes([0])
SEND = bytes([1])


class TelnetHandlerBase(AsyncBaseRequestHandler):
    """A telnet server based on the client in telnetlib"""

    # What I am prepared to do?
    DOACK = {
        ECHO: WILL,
        SGA: WILL,
        NEW_ENVIRON: WONT,
    }
    # What do I want the client to do?
    WILLACK = {
        ECHO: DONT,
        SGA: DO,
        NAWS: DONT,
        TTYPE: DO,
        LINEMODE: DONT,
        NEW_ENVIRON: DO,
    }
    # Default terminal type - used if client doesn't tell us its termtype
    TERM = "ansi"
    # Keycode to name mapping - used to decide which keys to query
    KEYS = {  # Key escape sequences
        curses.KEY_UP: 'Up',  # Cursor up
        curses.KEY_DOWN: 'Down',  # Cursor down
        curses.KEY_LEFT: 'Left',  # Cursor left
        curses.KEY_RIGHT: 'Right',  # Cursor right
        curses.KEY_DC: 'Delete',  # Delete right
        curses.KEY_BACKSPACE: 'Backspace',  # Delete left
    }
    # Reverse mapping of KEYS - used for cooking key codes
    ESCSEQ = {
    }
    # Terminal output escape sequences
    CODES = {
        'DEOL': b'',  # Delete to end of line
        'DEL': b'',  # Delete and close up
        'INS': b'',  # Insert space
        'CSRLEFT': b'',  # Move cursor left 1 space
        'CSRRIGHT': b'',  # Move cursor right 1 space
    }

    # --------------------------- Environment Setup ----------------------------

    def __init__(self, reader, writer, client_address, loop):
        """Constructor.

        When called without arguments, create an unconnected instance.
        With a hostname argument, it connects the instance; a port
        number is optional.
        """
        # Am I doing the echoing?
        self.loop = loop
        self.DOECHO = True
        # What opts have I sent DO/DONT for and what did I send?
        self.DOOPTS = {}
        # What opts have I sent WILL/WONT for and what did I send?
        self.WILLOPTS = {}
        self.reader = reader
        self.writer = writer
        # What commands does this CLI support
        self.rawq = b''  # Raw input string
        self.sbdataq = b''  # Sub-Neg string
        self.eof = 0  # Has EOF been reached?
        self.iacseq = b''  # Buffer for IAC sequence.
        self.sb = 0  # Flag for SB and SE sequence.
        self.history = []  # Command history
        self.cookedq = asyncio.Queue(loop=self.loop)
        super().__init__(reader, writer, client_address)

    def setterm(self, term):
        """Set the curses structures for this terminal"""
        raise NotImplementedError("Please Implement the setterm method")

    def setup(self):
        """Connect incoming connection to a telnet session"""
        self.setterm(self.TERM)
        for k in self.DOACK.keys():
            self.sendcommand(self.DOACK[k], k)
        for k in self.WILLACK.keys():
            self.sendcommand(self.WILLACK[k], k)

        self.inputcooker_task = asyncio.ensure_future(self.inputcooker(), loop=self.loop)

    def finish(self):
        """End this session"""
        log.debug("Session disconnected.")
        self.session_end()

    def session_start(self):
        pass

    def session_end(self):
        pass

    # ------------------------- Telnet Options Engine --------------------------

    def sendcommand(self, cmd, opt=None):
        """Send a telnet command (IAC)"""
        if cmd in [DO, DONT]:
            if opt not in self.DOOPTS:
                self.DOOPTS[opt] = None
            if (((cmd == DO) and (self.DOOPTS[opt] is not True))
                    or ((cmd == DONT) and (self.DOOPTS[opt] is not False))):
                self.DOOPTS[opt] = (cmd == DO)
                self.writecooked(IAC + cmd + opt)
        elif cmd in [WILL, WONT]:
            if opt not in self.WILLOPTS:
                self.WILLOPTS[opt] = b''
            if (((cmd == WILL) and (self.WILLOPTS[opt] is not True))
                    or ((cmd == WONT) and (self.WILLOPTS[opt] is not False))):
                self.WILLOPTS[opt] = (cmd == WILL)
                self.writecooked(IAC + cmd + opt)
        else:
            self.writecooked(IAC + cmd)

    # ---------------------------- Input Functions -----------------------------

    def _readline_do_echo(self, echo):
        """Determine if we should echo or not"""
        return echo is True or (echo is None and self.DOECHO is True)

    def _readline_echo(self, char, echo):
        """Echo a recieved character, move cursor etc..."""
        if self._readline_do_echo(echo):
            self.write(char)

    _current_line = b''
    _current_prompt = b''

    async def ansi_to_curses(self, char):
        """Handles reading ANSI escape sequences"""
        # ANSI sequences are:
        # ESC [ <key>
        # If we see ESC, read a char
        if char != 27:     # ESC = bytes([27])
            return char
        # If we see [, read another char
        if convert_to_bytes(await self.getc()) != ANSI_START_SEQ:
            self._readline_echo(BELL, True)
            return 0   # theNull = bytes([0])
        key = convert_to_bytes(await self.getc())
        # Translate the key to curses
        try:
            return ANSI_KEY_TO_CURSES[key]
        except KeyError:
            self._readline_echo(BELL, True)
            return 0

    def _readline_insert(self, charb, echo, insptr, line):
        """Deal properly with inserted chars in a line."""
        if not self._readline_do_echo(echo):
            return
        # Write out the remainder of the line
        self.write(charb + b''.join(line[insptr:]))
        # Cursor Left to the current insert point
        char_count = len(line) - insptr
        self.write(self.CODES['CSRLEFT'] * char_count)

    async def readline(self, echo=None, prompt=b'', use_history=True):
        """Return a line of bytes, including the terminating LF
           If echo is true always echo, if echo is false never echo
           If echo is None follow the negotiated setting.
           prompt is the current prompt to write (and rewrite if needed)
           use_history controls if this current line uses (and adds to) the command history.
        """

        line = []
        insptr = 0
        histptr = len(self.history)

        if self.DOECHO:
            self.write(prompt)
            self._current_prompt = prompt
        else:
            self._current_prompt = b''

        self._current_line = b''

        while True:
            c = await self.getc()
            c = await self.ansi_to_curses(c)
            cb = convert_to_bytes(c)

            if cb == theNULL:
                continue

            elif c == curses.KEY_LEFT:
                if insptr > 0:
                    insptr -= 1
                    self._readline_echo(self.CODES['CSRLEFT'], echo)
                else:
                    self._readline_echo(BELL, echo)
                continue
            elif c == curses.KEY_RIGHT:
                if insptr < len(line):
                    insptr += 1
                    self._readline_echo(self.CODES['CSRRIGHT'], echo)
                else:
                    self._readline_echo(BELL, echo)
                continue
            elif c == curses.KEY_UP or c == curses.KEY_DOWN:
                if not use_history:
                    self._readline_echo(BELL, echo)
                    continue
                if c == curses.KEY_UP:
                    if histptr > 0:
                        histptr -= 1
                    else:
                        self._readline_echo(BELL, echo)
                        continue
                elif c == curses.KEY_DOWN:
                    if histptr < len(self.history):
                        histptr += 1
                    else:
                        self._readline_echo(BELL, echo)
                        continue
                line = []
                if histptr < len(self.history):
                    line.extend(self.history[histptr])
                for _ in range(insptr):
                    self._readline_echo(self.CODES['CSRLEFT'], echo)
                self._readline_echo(self.CODES['DEOL'], echo)
                self._readline_echo(b''.join(line), echo)
                insptr = len(line)
                continue
            elif cb == bytes([3]):
                self._readline_echo(b'\n' + convert_to_bytes(curses.ascii.unctrl(c)) + b' ABORT\n', echo)
                return b''
            elif cb == bytes([4]):
                if len(line) > 0:
                    self._readline_echo(b'\n' + convert_to_bytes(curses.ascii.unctrl(c)) + b' ABORT (QUIT)\n', echo)
                    return b''
                self._readline_echo(b'\n' + convert_to_bytes(curses.ascii.unctrl(c)) + b' QUIT\n', echo)
                return b'QUIT'
            elif cb == bytes([10]):
                self._readline_echo(cb, echo)
                result = b''.join(convert_to_bytes(elem) for elem in line)
                if use_history:
                    self.history.append(result)
                if echo is False:
                    if prompt:
                        self.write(bytes([10]))
                    log.debug('readline: %s(hidden text)', prompt)
                else:
                    log.debug('readline: %s%r', prompt, result)
                return result
            elif c == curses.KEY_BACKSPACE or cb == bytes([127]) or cb == bytes([8]):
                if insptr > 0:
                    self._readline_echo(self.CODES['CSRLEFT'] + self.CODES['DEL'], echo)
                    insptr -= 1
                    del line[insptr]
                else:
                    self._readline_echo(BELL, echo)
                continue
            elif c == curses.KEY_DC:
                if insptr < len(line):
                    self._readline_echo(self.CODES['DEL'], echo)
                    del line[insptr]
                else:
                    self._readline_echo(BELL, echo)
                continue
            else:
                if c < 32:
                    c = curses.ascii.unctrl(c)
                    cb = convert_to_bytes(c)
                if len(line) > insptr:
                    self._readline_insert(cb, echo, insptr, line)
                else:
                    self._readline_echo(cb, echo)
            line[insptr:insptr] = cb
            insptr += len(cb)
            if self._readline_do_echo(echo):
                self._current_line = line

    async def getc(self):
        """Return one character from the input queue"""
        try:
            return await self.cookedq.get()
        except asyncio.QueueEmpty:
            return b''

    # --------------------------- Output Functions -----------------------------

    def write(self, data_bytes):
        """Send a packet. This function cooks output."""
        data_bytes = data_bytes.replace(IAC, IAC + IAC)
        data_bytes = data_bytes.replace(bytes([10]), bytes([13]) + bytes([10]))
        self.writecooked(data_bytes)

    def writecooked(self, data_bytes):
        """Put data directly into the output queue (bypass output cooker)"""
        self.writer.write(data_bytes)

    def writeline(self, data_bytes):
        """Send a packet with line ending."""
        log.debug('writing line %r' % data_bytes)
        self.write(data_bytes + bytes([10]))

    # ------------------------------- Input Cooker -----------------------------
    async def _inputcooker_getc(self):
        """Get one character from the raw queue.
        Raise EOFError on end of stream. SHOULD ONLY BE CALLED FROM THE
        INPUT COOKER."""
        if self.rawq:
            ret = self.rawq[0]
            self.rawq = self.rawq[1:]
            return bytes([ret]) if ret else b''
        try:
            ret = await self.reader.read(20)
        except BrokenPipeError:
            ret = b''

        self.eof = not ret
        self.rawq = self.rawq + ret
        if self.eof:
            raise EOFError
        return await self._inputcooker_getc()

    def _inputcooker_ungetc(self, char):
        """Put characters back onto the head of the rawq. SHOULD ONLY
        BE CALLED FROM THE INPUT COOKER."""
        self.rawq = char + self.rawq

    async def _inputcooker_store(self, char):
        """Put the cooked data in the correct queue"""
        if self.sb:
            self.sbdataq = self.sbdataq + char
        else:
            await self.inputcooker_store_queue(char)

    async def inputcooker_store_queue(self, char):
        """Put the cooked data in the input queue (no locking needed)"""
        if isinstance(char, list) or isinstance(char, tuple) \
                or isinstance(char, str) or isinstance(char, bytes):
            for v in char:
                await self.cookedq.put(v)
        else:
            await self.cookedq.put(char)

    async def inputcooker(self):
        """Input Cooker - Transfer from raw queue to cooked queue.

        Set self.eof when connection is closed.
        """
        try:
            while True:
                cb = await self._inputcooker_getc()
                if not self.iacseq:
                    if cb == IAC:
                        self.iacseq += cb
                        continue
                    elif cb == bytes([13]) and not self.sb:
                        c2b = await self._inputcooker_getc()
                        if c2b == theNULL or c2b == b'':
                            cb = bytes([10])
                        elif c2b == bytes([10]):
                            cb = c2b
                        else:
                            self._inputcooker_ungetc(c2b)
                            cb = bytes([10])
                    elif cb in [x[0] for x in self.ESCSEQ.keys()]:
                        # Looks like the begining of a key sequence
                        codes = cb
                        for keyseq in self.ESCSEQ.keys():
                            if len(keyseq) == 0:
                                continue
                            while codes == keyseq[:len(codes)] and len(codes) <= keyseq:
                                if codes == keyseq:
                                    cb = self.ESCSEQ[keyseq]
                                    break
                                codes = codes + await self._inputcooker_getc()
                            if codes == keyseq:
                                break
                            self._inputcooker_ungetc(codes[1:])
                            codes = codes[0]
                    await self._inputcooker_store(cb)
                elif len(self.iacseq) == 1:
                    # IAC: IAC CMD [OPTION only for WILL/WONT/DO/DONT]
                    if cb in (DO, DONT, WILL, WONT):
                        self.iacseq += cb
                        continue
                    self.iacseq = b''
                    if cb == IAC:
                        await self._inputcooker_store(cb)
                    else:
                        if cb == SB:  # SB ... SE start.
                            self.sb = 1
                            self.sbdataq = b''
                        elif cb == SE:  # SB ... SE end.
                            self.sb = 0
                        # Callback is supposed to look into
                        # the sbdataq
                elif len(self.iacseq) == 2:
                    self.iacseq = b''
        except EOFError:
            pass

    async def authentication_ok(self):
        """Checks the authentication and sets the username of the currently connected terminal. Returns True or False"""
        raise NotImplementedError("Please Implement the authentication_ok method")

    # ----------------------- Command Line Processor Engine --------------------

    async def handle(self):
        """The actual service to which the user has connected."""
        await self.authentication_ok()
        self.inputcooker_task.cancel()
        try:
            await self.inputcooker_task
        except asyncio.CancelledError:
            pass


def convert_to_bytes(c):
    if isinstance(c, int):
        if c < 256:
            cb = bytes([c])
        else:
            cb = None
    elif isinstance(c, str):
        cb = bytes(c, 'utf-8')
    else:
        cb = c
    return cb
# vim: set syntax=python ai showmatch:



# Copyright (C) 2013 Aniket Panse <contact@aniketpanse.in>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging
import binascii

from heralding.capabilities.handlerbase import HandlerBase

# VNC constants
RFB_VERSION = b'RFB 003.007\n'
AUTH_METHODS = b'\x01\x02'
VNC_AUTH = b'\x02'
AUTH_FAILED = b'\x00\x00\x00\x01'

logger = logging.getLogger(__name__)


class Vnc(HandlerBase):
    async def execute_capability(self, reader, writer, session):
        await self._handle_session(reader, writer, session)

    async def _handle_session(self, reader, writer, session):
        writer.write(RFB_VERSION)
        client_version = await reader.read(1024)

        if client_version == RFB_VERSION:
            await self.security_handshake(reader, writer, session)
        else:
            session.end_session()

    async def security_handshake(self, reader, writer, session):
        writer.write(AUTH_METHODS)
        sec_method = await reader.read(1024)

        if sec_method == VNC_AUTH:
            await self.do_vnc_authentication(reader, writer, session)
        else:
            session.end_session()

    async def do_vnc_authentication(self, reader, writer, session):
        challenge = os.urandom(16)
        writer.write(challenge)

        client_response = await reader.read(1024)
        session.add_auth_attempt('des_challenge', response=binascii.hexlify(client_response).decode(),
                                 challenge=binascii.hexlify(challenge).decode())

        writer.write(AUTH_FAILED)
        session.end_session()
